use common::CreateAchievement;
use common::CreateMilestone;
use common::DeleteAchievement;
use common::DeleteMilestone;
use common::RestResponse;
use common::ToggleAchievement;
use reqwasm::http::Request;
use serde::de::DeserializeOwned;
use serde::Serialize;

pub struct RestService;

#[derive(thiserror::Error, Debug)]
pub enum RestServiceError {
    #[error("Failed to serialize/deserialize data: {0}")]
    SerdeError(#[from] serde_json::Error),

    #[error(transparent)]
    RestError(#[from] common::RestError),

    #[error("Failed request: {0}")]
    ReqwasmError(#[from] reqwasm::Error),

    #[error("Unexpected response")]
    UnexpectedResponse(reqwasm::http::Response),
}

impl RestService {
    fn hostname() -> String {
        web_sys::window().unwrap().location().hostname().unwrap()
    }

    fn url_to(endpoint: &str) -> String {
        format!("http://{}:4000{endpoint}", Self::hostname())
    }

    async fn post_json<P: Serialize, T: DeserializeOwned>(
        payload: P,
        path: &str,
    ) -> Result<T, RestServiceError> {
        let req = Request::post(&Self::url_to(path))
            .header("Content-Type", "application/json")
            .body(serde_json::to_string(&payload)?);

        let response = req.send().await.map_err(|err| {
            log::error!("Request failed: {:?}", err);
            err
        })?;

        if let Ok(rest_response) = response.json::<RestResponse<T>>().await {
            return rest_response.map_err(Into::into);
        }

        let err = RestServiceError::UnexpectedResponse(response);
        log::warn!("{err:?}");
        Err(err)
    }

    pub async fn toggle_achievement(payload: ToggleAchievement) -> Result<(), RestServiceError> {
        Self::post_json(payload, "/toggle").await
    }

    pub async fn create_achievement(payload: CreateAchievement) -> Result<(), RestServiceError> {
        Self::post_json(payload, "/create").await
    }

    pub async fn delete_achievement(payload: DeleteAchievement) -> Result<(), RestServiceError> {
        Self::post_json(payload, "/delete").await
    }

    pub async fn create_milestone(payload: CreateMilestone) -> Result<(), RestServiceError> {
        Self::post_json(payload, "/create-milestone").await
    }

    pub async fn delete_milestone(payload: DeleteMilestone) -> Result<(), RestServiceError> {
        Self::post_json(payload, "/delete-milestone").await
    }
}
